/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007-2014 Michael Cornelison
   Source URL: http://kornelix.com/fotoxx
   Contact: kornelix@posteo.de
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"

/**************************************************************************
   Fotoxx image edit - Retouch menu functions
***************************************************************************/


//  automatic image tuneup without user guidance

float       voodoo_brdist[256];                                            //  pixel count per brightness level
int         voodoo_01, voodoo_99;                                          //  1% and 99% brightness levels (0 - 255)
void        voodoo_distribution();                                         //  compute brightness distribution
void *      voodoo_thread(void *);
editfunc    EFvoodoo;


void m_voodoo(GtkWidget *, cchar *)                                        //  new v.12.06
{
   F1_help_topic = "voodoo";

   EFvoodoo.menufunc = m_voodoo;
   EFvoodoo.funcname = "voodoo";
   EFvoodoo.Farea = 1;                                                     //  select area ignored
   EFvoodoo.threadfunc = voodoo_thread;
   if (! edit_setup(EFvoodoo)) return;                                     //  setup edit
   
   voodoo_distribution();                                                  //  compute brightness distribution
   signal_thread();                                                        //  start update thread
   edit_done(0);                                                           //  edit done
   return;
}


//  compute brightness distribution for image

void voodoo_distribution()
{
   int         px, py, ii;
   int         ww = E1pxm->ww, hh = E1pxm->hh;
   float       bright1;
   float       *pix1;

   for (ii = 0; ii < 256; ii++)                                            //  clear brightness distribution data
      voodoo_brdist[ii] = 0;

   for (py = 0; py < hh; py++)                                             //  compute brightness distribution
   for (px = 0; px < ww; px++)
   {
      pix1 = PXMpix(E1pxm,px,py);
      bright1 = pixbright(pix1);
      voodoo_brdist[int(bright1)]++;
   }

   for (ii = 1; ii < 256; ii++)                                            //  cumulative brightness distribution
      voodoo_brdist[ii] += voodoo_brdist[ii-1];                            //   0 ... (ww * hh)
   
   voodoo_01 = 0;
   voodoo_99 = 255;

   for (ii = 0; ii < 256; ii++)                                            //  compute index values (0 - 255)
   {                                                                       //    for darkest 1% and brightest 1%
      if (voodoo_brdist[ii] < 0.01 * ww * hh) voodoo_01 = ii;
      if (voodoo_brdist[ii] < 0.99 * ww * hh) voodoo_99 = ii;
   }
   
   for (ii = 0; ii < 256; ii++)
      voodoo_brdist[ii] = voodoo_brdist[ii]                                //  multiplier per brightness level 
                    / (ww * hh) * 256.0 / (ii + 1);                        //  ( = 1 for a flat distribution)
   return;
}


//  thread function - use multiple working threads

void * voodoo_thread(void *)
{
   void  * voodoo_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(voodoo_wthread,&Nval[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmods++; 
      CEF->Fsaved = 0;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}

void * voodoo_wthread(void *arg)                                           //  worker thread function
{
   int         index = *((int *) (arg));
   int         px, py;
   float       *pix1, *pix3;
   float       bright1, bright2, bright3, cmax;
   float       red1, green1, blue1, red3, green3, blue3;
   float       flat1 = 0.3;                                                //  strength of distribution flatten
   float       flat2 = 1.0 - flat1;
   float       sat1 = 0.4, sat2;                                           //  strength of saturation increase
   float       f1, f2;
   float       expand = 256.0 / (voodoo_99 - voodoo_01 + 1);               //  brightness range expander
   
   for (py = index; py < E1pxm->hh; py += Nwt)                             //  voodoo brightness distribution
   for (px = 0; px < E1pxm->ww; px++)
   {
      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel

      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];

      bright2 = 0.25 * red1 + 0.65 * green1 + 0.10 * blue1;                //  input brightness, 0 - 256
      bright2 = (bright2 - voodoo_01) * expand;                            //  expand to clip low / high 1% 
      if (bright2 < 0) bright2 = 0;
      if (bright2 > 255) bright2 = 255;

      bright1 = voodoo_brdist[int(bright2)];                               //  factor for flat output brightness
      bright3 = flat1 * bright1 + flat2;                                   //  attenuate

      f1 = (256.0 - bright2) / 256.0;                                      //  bright2 = 0 - 256  >>  f1 = 1 - 0
      f2 = 1.0 - f1;                                                       //  prevent banding in bright areas
      bright3 = f1 * bright3 + f2;                                         //  tends to 1.0 for brighter pixels

      red3 = red1 * bright3;                                               //  blend new and old brightness
      green3 = green1 * bright3;
      blue3 = blue1 * bright3;
      
      bright3 = 0.333 * (red3 + green3 + blue3);                           //  mean color brightness
      sat2 = sat1 * (256.0 - bright3) / 256.0;                             //  bright3 = 0 - 256  >>  sat2 = sat1 - 0
      red3 = red3 + sat2 * (red3 - bright3);                               //  amplified color, max for dark pixels
      green3 = green3 + sat2 * (green3 - bright3);
      blue3 = blue3 + sat2 * (blue3 - bright3);
      
      if (red3 < 0) red3 = 0;                                              //  stop possible underflow
      if (green3 < 0) green3 = 0;
      if (blue3 < 0) blue3 = 0;

      cmax = red3;                                                         //  stop overflow, keep color balance
      if (green3 > cmax) cmax = green3;
      if (blue3 > cmax) cmax = blue3;
      if (cmax > 255.9) {
         cmax = 255.9 / cmax;
         red3 = red3 * cmax;
         green3 = green3 * cmax;
         blue3 = blue3 * cmax;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  Retouch Combo
//  Adjust all aspects of brightness, contrast, color.
//  Brightness curves are used, overall and per RGB color.
//  Color saturation and white balance adjustments are also available.

namespace combo_names
{
   editfunc    EFcombo;
   int         combo_spc;                                                  //  current spline curve
   float       brightness;                                                 //  brightness input, -1 ... +1
   float       contrast;                                                   //  contrast input, -1 ... +1
   float       wbalR, wbalG, wbalB;                                        //  white balance, white basis standard
   int         wbtemp;                                                     //  illumination temp. input, 1K ... 10K
   float       tempR, tempG, tempB;                                        //  RGB of illumination temp.
   float       combR, combG, combB;                                        //  combined RGB factors
   float       satlevel;                                                   //  saturation input, -1 ... +1 = saturated
   float       areaemph;                                                   //  emphasis input, -1 ... +1 = bright areas
   float       emphcurve[256];                                             //  emphasis per brightness level, 0 ... 1
   int         Fapply = 0;                                                 //  flag, apply dialog controls to image
   char        combo_curves[200] = "";                                     //  previous curves save file
}

void  blackbodyRGB(int K, float &R, float &G, float &B);


//  menu function

void m_combo(GtkWidget *, cchar *)
{
   using namespace combo_names;

   int    combo_dialog_event(zdialog* zd, cchar *event);
   void   combo_curvedit(int spc);
   void * combo_thread(void *);

   GtkWidget   *drawareaB, *frameB;
   
   F1_help_topic = "retouch_combo";

   if (! *combo_curves)                                                    //  initz. previous curves save file
      snprintf(combo_curves,200,"%s/RCprev.curve",saved_curves_dirk);

   EFcombo.menufunc = m_combo;
   EFcombo.funcname = "retouch_combo";
   EFcombo.FprevReq = 1;                                                   //  use preview
   EFcombo.Farea = 2;                                                      //  select area usable
   EFcombo.Frestart = 1;                                                   //  restart allowed
   EFcombo.threadfunc = combo_thread;
   if (! edit_setup(EFcombo)) return;                                      //  setup edit
   
/***
       _____________________________________________________
      |                    Retouch Combo                    |
      |  _________________________________________________  |
      | |                                                 | |              //  5 curves are maintained:
      | |                                                 | |              //  curve 0: current display curve
      | |                                                 | |              //        1: curve for all colors
      | |         brightness curve area                   | |              //        2,3,4: red, green, blue
      | |                                                 | |
      | |                                                 | |
      | |_________________________________________________| |
      | |_________________________________________________| |              //  brightness scale: black to white stripe
      |   (o) all  (o) red  (o) green  (o) blue             |              //  select curve to display
      |   Brightness ================[]=============  High  |              //  brightness
      |    Contrast  ================[]=============  High  |              //  contrast
      |    Low Color ====================[]=========  High  |              //  color saturation
      |     Warmer   ====================[]========= Cooler |              //  color temperature
      |   Dark Areas ==========[]=================== Bright |              //  color emphasis
      |  [x] Click for white balance or black level         |
      |                                                     |              //  click gray/white spot for white balance
      |              [Bdist] [Reset] [Prev] [Done] [Cancel] |
      |_____________________________________________________|

***/

   EFcombo.zd = zdialog_new(ZTX("Retouch Combo"),mWin,Bdist,Breset,Bprev,Bdone,Bcancel,null);
   zdialog *zd = EFcombo.zd;

   zdialog_add_widget(zd,"frame","frame","dialog",0,"expand");             //  curve drawing area

   zdialog_add_widget(zd,"frame","frameB","dialog");                       //  black to white brightness scale
   frameB = zdialog_widget(zd,"frameB");
   drawareaB = gtk_drawing_area_new();
   gtk_container_add(GTK_CONTAINER(frameB),drawareaB);
   gtk_widget_set_size_request(drawareaB,300,12);
   G_SIGNAL(drawareaB,"draw",draw_brightscale,0);

   zdialog_add_widget(zd,"hbox","hbrgb","dialog",0,"space=3");
   zdialog_add_widget(zd,"radio","all","hbrgb",Ball,"space=5");
   zdialog_add_widget(zd,"radio","red","hbrgb",Bred,"space=3");
   zdialog_add_widget(zd,"radio","green","hbrgb",Bgreen,"space=3");
   zdialog_add_widget(zd,"radio","blue","hbrgb",Bblue,"space=3");
   
   zdialog_add_widget(zd,"hbox","hbcolor","dialog");
   zdialog_add_widget(zd,"label","space","hbcolor",0,"space=5");
   zdialog_add_widget(zd,"vbox","vbcolor1","hbcolor",0,"homog");
   zdialog_add_widget(zd,"label","space","hbcolor",0,"space=3");
   zdialog_add_widget(zd,"vbox","vbcolor2","hbcolor",0,"homog|expand");
   zdialog_add_widget(zd,"label","space","hbcolor",0,"space=3");
   zdialog_add_widget(zd,"vbox","vbcolor3","hbcolor",0,"homog");
   zdialog_add_widget(zd,"label","space","hbcolor",0,"space=5");

   zdialog_add_widget(zd,"label","labrite","vbcolor1",ZTX("Brightness"));
   zdialog_add_widget(zd,"label","labcont","vbcolor1",ZTX("Contrast"));
   zdialog_add_widget(zd,"label","labsat1","vbcolor1",ZTX("Low Color"));
   zdialog_add_widget(zd,"label","labtemp1","vbcolor1",ZTX("Warmer"));
   zdialog_add_widget(zd,"label","labarea1","vbcolor1",ZTX("Dark Areas"));

   zdialog_add_widget(zd,"hscale","brightness","vbcolor2","-1.0|1.0|0.002|0");
   zdialog_add_widget(zd,"hscale","contrast","vbcolor2","-1.0|1.0|0.002|0");
   zdialog_add_widget(zd,"hscale","satlevel","vbcolor2","-1.0|1.0|0.002|0");
   zdialog_add_widget(zd,"hscale","wbtemp","vbcolor2","1000|9999|1|5000");
   zdialog_add_widget(zd,"hscale","areaemph","vbcolor2","0|1.0|0.002|0.5");

   zdialog_add_widget(zd,"label","labrite2","vbcolor3",ZTX("High"));
   zdialog_add_widget(zd,"label","labcont2","vbcolor3",ZTX("High"));
   zdialog_add_widget(zd,"label","labsat2","vbcolor3",ZTX("High"));
   zdialog_add_widget(zd,"label","labtemp2","vbcolor3",ZTX("Cooler"));
   zdialog_add_widget(zd,"label","labarea2","vbcolor3",ZTX("Bright"));
   
   zdialog_add_widget(zd,"hbox","hbclick","dialog",0,"space=3");
   zdialog_add_widget(zd,"check","click","hbclick",ZTX("Click for white balance or black level"),"space=3");
   
   GtkWidget *frame = zdialog_widget(zd,"frame");
   spldat *sd = splcurve_init(frame,combo_curvedit);
   EFcombo.curves = sd;

   sd->Nscale = 1;                                                         //  diagonal fixed line, neutral curve
   sd->xscale[0][0] = 0.00;
   sd->yscale[0][0] = 0.00;
   sd->xscale[1][0] = 1.00;
   sd->yscale[1][0] = 1.00;

   for (int ii = 0; ii < 5; ii++)                                          //  loop curves 0-4
   {
      sd->nap[ii] = 5;                                                     //  initial curves are neutral
      sd->vert[ii] = 0;
      sd->apx[ii][0] = sd->apy[ii][0] = 0.01;
      sd->apx[ii][1] = sd->apy[ii][1] = 0.25;
      sd->apx[ii][2] = sd->apy[ii][2] = 0.50;
      sd->apx[ii][3] = sd->apy[ii][3] = 0.75;
      sd->apx[ii][4] = sd->apy[ii][4] = 0.99;
      splcurve_generate(sd,ii);
   }

   sd->Nspc = 1;                                                           //  only curve 0 is shown 
   combo_spc = 1;                                                          //  current active curve: 1 (all)
   zdialog_stuff(zd,"all",1);                                              //  stuff default selection, all

   brightness = 0;                                                         //  neutral brightness                 v.14.02
   contrast = 0;                                                           //  neutral contrast                   v.14.02

   wbalR = wbalG = wbalB = 1.0;                                            //  neutral white balance 
   wbtemp = 5000;                                                          //  neutral illumination temp.
   blackbodyRGB(wbtemp,tempR,tempG,tempB);                                 //  all are 1.0
   combR = combG = combB = 1.0;                                            //  neutral RGB factors

   satlevel = 0;                                                           //  neutral saturation
   
   areaemph = 0.5;                                                         //  even dark/bright area emphasis
   for (int ii = 0; ii < 256; ii++)
      emphcurve[ii] = 1;
   
   zdialog_stuff(zd,"click",0);                                            //  reset mouse click status

   zdialog_resize(zd,0,350);
   zdialog_run(zd,combo_dialog_event,"save");                              //  run dialog - parallel
   
   return;
}


//  dialog event and completion callback function

int combo_dialog_event(zdialog *zd, cchar *event)
{
   using namespace combo_names;

   void  combo_mousefunc();

   spldat      *sd = EFcombo.curves;
   float       c1, c2, c3, xval, yval;
   float       bright0, dbrite, cont0, dcont;
   float       dx, dy;
   int         ii, jj;
   int         Fapply = 0;
   
   if (strEqu(event,"done")) zd->zstat = 4;                                //  apply and quit 
   if (strEqu(event,"enter")) zd->zstat = 4;                               //  v.14.03
   
   if (zd->zstat == 1) {                                                   //  [Bdist]
      zd->zstat = 0;                                                       //  keep dialog active
      m_histogram(0,0);                                                    //  show brightness distribution
   }
   
   if (zd->zstat == 2)                                                     //  [reset]
   {
      zd->zstat = 0;                                                       //  keep dialog active
      
      zdialog_stuff(zd,"brightness",0);                                    //  neutral brightness                 v.14.02
      zdialog_stuff(zd,"contrast",0);                                      //  neutral contrast                   v.14.02
      brightness = contrast = 0;

      wbalR = wbalG = wbalB = 1.0;                                         //  neutral white balance
      wbtemp = 5000;                                                       //  neutral illumination temp.
      zdialog_stuff(zd,"wbtemp",wbtemp);
      blackbodyRGB(wbtemp,tempR,tempG,tempB);                              //  all are 1.0
      combR = combG = combB = 1.0;                                         //  neutral RGB factors
      
      satlevel = 0;                                                        //  neutral saturation
      zdialog_stuff(zd,"satlevel",0);

      areaemph = 0.5;                                                      //  even area emphasis
      zdialog_stuff(zd,"areaemph",0.5);
      for (int ii = 0; ii < 256; ii++)                                     //  flat curve
         emphcurve[ii] = 1;

      for (int ii = 0; ii < 5; ii++)                                       //  loop gamma curves 0-4
      {
         sd->nap[ii] = 5;                                                  //  all curves are neutral
         sd->vert[ii] = 0;
         sd->apx[ii][0] = sd->apy[ii][0] = 0.01;
         sd->apx[ii][1] = sd->apy[ii][1] = 0.25;
         sd->apx[ii][2] = sd->apy[ii][2] = 0.50;
         sd->apx[ii][3] = sd->apy[ii][3] = 0.75;
         sd->apx[ii][4] = sd->apy[ii][4] = 0.99;
         splcurve_generate(sd,ii);
      }

      combo_spc = 1;                                                       //  current active curve: 1 (all)
      zdialog_stuff(zd,"all",1);
      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve
      
      edit_reset();                                                        //  restore initial image 
      return 1;
   }
   
   if (zd->zstat == 3)                                                     //  [prev] restore previous settings
   {
      zd->zstat = 0;                                                       //  keep dialog active
      zdialog_restore_inputs(zd);                                          //  restore previous inputs

      sd->Nspc = 5;                                                        //  5 curves
      splcurve_load(sd,combo_curves);                                      //  load all curves from file
      sd->Nspc = 1;                                                        //  only one curve active
      for (int ii = 0; ii < 5; ii++)
         splcurve_generate(sd,ii);                                         //  v.14.02
      Fapply = 1;
   }

   if (zd->zstat)                                                          //  [done] or [cancel]
   {
      m_histogram(0,"kill");
      freeMouse();
      if (zd->zstat == 4) {                                                //  [done]
         sd->Nspc = 5;                                                     //  include all 5 curves
         splcurve_save(sd,combo_curves);                                   //  save curves for next time
         edit_done(0);                                                     //  saves zdialog inputs
      }
      else edit_cancel(0);                                                 //  [cancel] or [x]
      return 1;
   }
   
   if (strstr("all red green blue",event))                                 //  new choice of curve
   {
      zdialog_fetch(zd,event,ii);
      if (! ii) return 0;                                                  //  ignore button OFF events

      ii = strcmpv(event,"all","red","green","blue",null);
      combo_spc = ii;                                                      //  1, 2, 3, 4

      sd->nap[0] = sd->nap[ii];                                            //  copy active curve ii to curve 0
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }

      splcurve_generate(sd,0);                                             //  regenerate curve 0
      
      for (jj = 0; jj < 1000; jj++)                                        //  copy curve values back
         sd->yval[ii][jj] = sd->yval[0][jj];

      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve

      if (combo_spc == 1) {                                                //  "all" curve (1) was selected
         for (ii = 2; ii < 5; ii++) {                                      //  fix R/G/B curves (2-4) to match
            sd->nap[ii] = sd->nap[1];
            for (jj = 0; jj < sd->nap[1]; jj++) {
               sd->apx[ii][jj] = sd->apx[1][jj];
               sd->apy[ii][jj] = sd->apy[1][jj];
            }

            for (jj = 0; jj < 1000; jj++)
               sd->yval[ii][jj] = sd->yval[1][jj];
         }
      }
   }
   
   if (strEqu(event,"brightness"))                                         //  brightness slider, 0 ... 1         v.14.02
   {
      bright0 = brightness;
      zdialog_fetch(zd,"brightness",brightness);
      dbrite = brightness - bright0;

      for (int ii = 0; ii < 5; ii++)                                       //  loop brightness curves 0-4
      {
         for (int jj = 0; jj < sd->nap[ii]; jj++)
         {
            dx = sd->apx[ii][jj];                                          //  0 ... 0.5 ... 1
            if (dx <= 0.01 || dx >= 0.99) continue;
            dx = 0.5 - fabsf(dx - 0.5);                                    //  0 ... 0.5 ... 0
            dx = 0.5 * sqrtf(dx);                                          //  0 ... 0.35 ... 0  (more rounded shape)
            dy = dx * dbrite;
            dy += sd->apy[ii][jj];
            if (dy < 0) dy = 0;
            if (dy > 1) dy = 1;
            sd->apy[ii][jj] = dy;
         }
         splcurve_generate(sd,ii);
      }

      combo_spc = 1;                                                       //  current active curve: 1 (all)
      zdialog_stuff(zd,"all",1);
      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve
   }
   
   if (strEqu(event,"contrast"))                                           //  contrast slider, 0 ... 1           v.14.02
   {
      cont0 = contrast;
      zdialog_fetch(zd,"contrast",contrast);
      dcont = contrast - cont0;

      for (int ii = 0; ii < 5; ii++)                                       //  loop brightness curves 0-4
      {
         for (int jj = 0; jj < sd->nap[ii]; jj++)
         {
            dx = sd->apx[ii][jj];                                          //  0 ... 0.5 ... 1
            if (dx <= 0.01 || dx >= 0.99) continue;
            if (dx < 0.5) 
               dx = -0.25 + fabsf(dx - 0.25);                              //  0 ... -0.25 ... 0
            else
               dx = +0.25 - fabsf(dx - 0.75);                              //  0 ... +0.25 ... 0
            dy = dx * dcont;
            dy += sd->apy[ii][jj];
            if (dy < 0) dy = 0;
            if (dy > 1) dy = 1;
            sd->apy[ii][jj] = dy;
         }
         splcurve_generate(sd,ii);
      }

      combo_spc = 1;                                                       //  current active curve: 1 (all)
      zdialog_stuff(zd,"all",1);
      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve
   }
   
   if (strEqu(event,"click")) {                                            //  toggle mouse click input
      zdialog_fetch(zd,"click",ii);
      if (ii) takeMouse(combo_mousefunc,dragcursor);                       //  on: connect mouse function
      else freeMouse();                                                    //  off: free mouse
   }

   if (strEqu(event,"blendwidth")) signal_thread();                        //  adjust area edge blending

   if (strstr("brightness contrast satlevel areaemph wbtemp",event))       //  dialog control changed
      Fapply = 1;
   
   if (! Fapply) return 1;
   
   zdialog_fetch(zd,"brightness",brightness);                              //  get brightness setting
   zdialog_fetch(zd,"contrast",contrast);                                  //  get contrast setting
   zdialog_fetch(zd,"satlevel",satlevel);                                  //  get saturation setting

   zdialog_fetch(zd,"wbtemp",wbtemp);                                      //  get illumination temp. setting
   blackbodyRGB(wbtemp,tempR,tempG,tempB);                                 //  generate new temp. adjustments
   
   zdialog_fetch(zd,"areaemph",areaemph);                                  //  get dark/bright area emphasis
      
   c1 = 2 * (1 - areaemph);                                                //  generate area emphasis curve
   if (c1 > 1) c1 = 1;
   c3 = 2 * areaemph;
   if (c3 > 1) c3 = 1;
   c2 = 0.7 * (c1 + c3);
   
   for (ii = 0; ii < 256; ii++) {
      xval = ii / 256.0;
      if (xval <= 0.5)  yval = c1 + 4 * (c2 - c1) * xval * xval;
      else  yval = c3 + 4 * (c2 - c3) * (1 - xval) * (1 - xval);
      if (yval > 1.0) yval = 1.0;
      emphcurve[ii] = yval;
   }

   signal_thread();
   return 1;
}


//  this function is called when a curve is edited

void combo_curvedit(int spc)
{
   using namespace combo_names;

   int      ii = combo_spc, jj;
   spldat   *sd = EFcombo.curves;                                          //  active curve, 1-4
   
   sd->nap[ii] = sd->nap[0];                                               //  copy curve 0 to active curve
   for (jj = 0; jj < sd->nap[0]; jj++) {
      sd->apx[ii][jj] = sd->apx[0][jj];
      sd->apy[ii][jj] = sd->apy[0][jj];
   }

   for (jj = 0; jj < 1000; jj++) 
      sd->yval[ii][jj] = sd->yval[0][jj];

   if (combo_spc == 1) {                                                   //  "all" curve (1) was edited
      for (ii = 2; ii < 5; ii++) {                                         //  fix R/G/B curves (2-4) to match
         sd->nap[ii] = sd->nap[1];
         for (jj = 0; jj < sd->nap[1]; jj++) {
            sd->apx[ii][jj] = sd->apx[1][jj];
            sd->apy[ii][jj] = sd->apy[1][jj];
         }

         for (jj = 0; jj < 1000; jj++)
            sd->yval[ii][jj] = sd->yval[1][jj];
      }
   }

   signal_thread();
   return;
}


//  get nominal white color from mouse click position

void combo_mousefunc()                                                     //  mouse function
{
   using namespace combo_names;

   int         px, py, dx, dy, ii, jj, err;
   float       red, green, blue, rgbmean;
   float       *ppix;
   spldat      *sd = EFcombo.curves;                                       //  active curve, 1-4
   zdialog     *zd = EFcombo.zd;
   char        mousetext[60];
   GdkScreen   *screen;
   GdkDevice   *mouse;

   if (! LMclick) return;
   LMclick = 0;
   
   px = Mxclick;                                                           //  mouse click position
   py = Myclick;
   
   if (px < 2) px = 2;                                                     //  pull back from edge
   if (px > E3pxm->ww-3) px = E3pxm->ww-3;
   if (py < 2) py = 2;
   if (py > E3pxm->hh-3) py = E3pxm->hh-3;
   
   red = green = blue = 0;

   for (dy = -1; dy <= 1; dy++)                                            //  3x3 block around mouse position
   for (dx = -1; dx <= 1; dx++)
   {
      ppix = PXMpix(E1pxm,px+dx,py+dy);                                    //  input image
      red += ppix[0];
      green += ppix[1];
      blue += ppix[2];
   }

   red = red / 9.0;                                                        //  mean RGB levels 
   green = green / 9.0;
   blue = blue / 9.0;

   snprintf(mousetext,60,"3x3 pixels RGB: %.0f %.0f %.0f \n",red,green,blue);
   err = get_mouse_device(mWin,&screen,&mouse);
   if (err) return;
   poptext_mouse(mousetext,mouse,10,10,0,3);
   
   if (red > 80 && green > 80 && blue > 80)                                //  assume this is a gray/white point
   {
      rgbmean = (red + green + blue) / 3.0;
      wbalR = rgbmean / red;                                               //  = 1.0 if all RGB are equal
      wbalG = rgbmean / green;
      wbalB = rgbmean / blue;
      signal_thread();                                                     //  trigger image update
   }
   
   else if (red < 50 && green < 50 && blue < 50)                           //  assume this is a black point
   {
      combo_spc = 1;                                                       //  curve 1 (all) is active curve
      zdialog_stuff(zd,"all",1);

      sd->apx[1][0] = (red + green + blue) / 3.0 / 256.0;                  //  set gamma curve low clipping level
      sd->apy[1][0] = 0.0;                                                 //    from black point pixel

      sd->nap[0] = sd->nap[1];                                             //  copy curve 1 to curve 0
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[1][jj];
         sd->apy[0][jj] = sd->apy[1][jj];
      }

      splcurve_generate(sd,0);                                             //  regenerate curve 0
      
      for (jj = 0; jj < 1000; jj++)                                        //  copy curve values back
         sd->yval[1][jj] = sd->yval[0][jj];

      for (ii = 2; ii < 5; ii++) {                                         //  fix R/G/B curves (2-4)
         sd->nap[ii] = sd->nap[1];                                         //    to match curve 1 (all)
         for (jj = 0; jj < sd->nap[1]; jj++) {
            sd->apx[ii][jj] = sd->apx[1][jj];
            sd->apy[ii][jj] = sd->apy[1][jj];
         }

         for (jj = 0; jj < 1000; jj++)
            sd->yval[ii][jj] = sd->yval[1][jj];
      }
      
      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve
      signal_thread();                                                     //  update image
   }

   return;
}


//  combo thread function

void * combo_thread(void *arg)
{
   using namespace combo_names;

   void * combo_wthread(void *);
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      combR = wbalR * tempR / 256.0;                                       //  gray standard based on 
      combG = wbalG * tempG / 256.0;                                       //    clicked pixel colors
      combB = wbalB * tempB / 256.0;                                       //      and illumination temp.

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(combo_wthread,&Nval[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      mwpaint2();                                                          //  update window
      CEF->Fmods++;                                                        //  image3 modified
      CEF->Fsaved = 0;
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * combo_wthread(void *arg)                                            //  worker thread function
{
   using namespace combo_names;
   
   int         index = *((int *) arg);
   int         ii, dist = 0, px, py;
   float       *pix1, *pix3;
   float       red1, green1, blue1, maxrgb;
   float       red3, green3, blue3;
   float       pixbrite, F1, F2;
   float       coeff = 1000.0 / 256.0;
   float       dold, dnew;
   spldat      *sd = EFcombo.curves;

   for (py = index; py < E3pxm->hh; py += Nwt)                             //  loop all pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      if (sa_stat == 3) {                                                  //  select area active
         ii = py * E3pxm->ww + px;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
      }

      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel

      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];

      //  apply white balance and temperature color shift

      red3 = combR * red1;
      green3 = combG * green1;
      blue3 = combB * blue1;

      maxrgb = red3;                                                       //  stop overflows
      if (green3 > maxrgb) maxrgb = green3;
      if (blue3 > maxrgb) maxrgb = blue3;

      if (maxrgb > 255.9) {
         red3 = red3 * 255.9 / maxrgb;
         green3 = green3 * 255.9 / maxrgb;
         blue3 = blue3 * 255.9 / maxrgb;
         maxrgb = 255.9;
      }

      //  apply saturation color shift

      pixbrite = 0.333 * (red3 + green3 + blue3);                          //  pixel brightness, 0 to 255.9
      red3 = red3 + satlevel * (red3 - pixbrite);                          //  satlevel is -1 ... +1
      green3 = green3 + satlevel * (green3 - pixbrite);
      blue3 = blue3 + satlevel * (blue3 - pixbrite);

      if (red3 < 0) red3 = 0;                                              //  stop underflow and overflow
      if (green3 < 0) green3 = 0;
      if (blue3 < 0) blue3 = 0;
      if (red3 > 255.9) red3 = 255.9;
      if (green3 > 255.9) green3 = 255.9;
      if (blue3 > 255.9) blue3 = 255.9;
      
      //  apply dark/bright area emphasis curve for color changes
      
      F1 = emphcurve[int(maxrgb)];                                         //  0 ... 1
      if (F1 < 1.0) {
         F2 = 1.0 - F1;
         red3 = F1 * red3 + F2 * red1;
         green3 = F1 * green3 + F2 * green1;
         blue3 = F1 * blue3 + F2 * blue1;
      }

      //  apply brightness shift

      ii = coeff * red3;                                                   //  range 0-999 for curve index
      red3 = 255.9 * sd->yval[2][ii];                                      //  output RGB brightness, 0-255.9
      ii = coeff * green3;
      green3 = 255.9 * sd->yval[3][ii];
      ii = coeff * blue3;
      blue3 = 255.9 * sd->yval[4][ii];
      
      maxrgb = red3;                                                       //  stop overflows
      if (green3 > maxrgb) maxrgb = green3;
      if (blue3 > maxrgb) maxrgb = blue3;

      if (maxrgb > 255.9) {
         red3 = red3 * 255.9 / maxrgb;
         green3 = green3 * 255.9 / maxrgb;
         blue3 = blue3 * 255.9 / maxrgb;
      }
      
      //  select area edge blending

      if (sa_stat == 3 && dist < sa_blend) {
         dnew = 1.0 * dist / sa_blend;
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  Return relative RGB illumination values for a light source
//     having a given input temperature of 1000-10000 deg. K
//  5000 K is neutral: all factors = 1.0

void blackbodyRGB(int K, float &R, float &G, float &B)
{
   float    kk[19] = { 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0, 6.5, 7.0, 7.5, 8.0, 8.5, 9.0, 9.5, 10.0 };
   float    r1[19] = { 255, 255, 255, 255, 255, 255, 255, 255, 254, 250, 242, 231, 220, 211, 204, 197, 192, 188, 184 };
   float    g1[19] = { 060, 148, 193, 216, 232, 242, 249, 252, 254, 254, 251, 245, 239, 233, 228, 224, 220, 217, 215 };
   float    b1[19] = { 000, 010, 056, 112, 158, 192, 219, 241, 253, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 };

   static int     ftf = 1;
   static float   r2[10000], g2[10000], b2[10000];

   if (ftf) {                                                              //  initialize
      spline1(19,kk,r1);
      for (int T = 1000; T < 10000; T++)
         r2[T] = spline2(0.001 * T);
      
      spline1(19,kk,g1);
      for (int T = 1000; T < 10000; T++)
         g2[T] = spline2(0.001 * T);
      
      spline1(19,kk,b1);
      for (int T = 1000; T < 10000; T++)
         b2[T] = spline2(0.001 * T);
      
      ftf = 0;
   }
   
   if (K < 1000 || K > 9999) zappcrash("blackbody bad K: %dK",K);
   
   R = r2[K];
   G = g2[K];
   B = b2[K];

   return;
}


/**************************************************************************

   Image Tone Mapping function
   enhance local contrast as opposed to overall contrast

   methodology:
   get brightness gradients for each pixel in 4 directions: SE SW NE NW
   amplify gradients using the edit curve (x-axis range 0-max. gradient)
   integrate 4 new brightness surfaces from the amplified gradients:
     - pixel brightness = prior pixel brightness + amplified gradient
     - the Amplify control varies amplification from zero to overamplified
   new pixel brightness = average from 4 calculated brightness surfaces

***************************************************************************/

float    *Tmap_brmap1;
float    *Tmap_brmap3[4];
int      Tmap_contrast99;
float    Tmap_amplify;

editfunc    EFtonemap;

void     Tmap_initz(zdialog *zd);
int      Tmap_dialog_event(zdialog *zd, cchar *event);
void     Tmap_curvedit(int);
void *   Tmap_thread(void *);


void m_tonemap(GtkWidget *, cchar *)
{
   F1_help_topic = "tone_mapping";

   cchar    *title = ZTX("Tone Mapping");

   EFtonemap.menufunc = m_tonemap;
   EFtonemap.funcname = "tonemap";
   EFtonemap.Farea = 2;                                                    //  select area usable
   EFtonemap.Frestart = 1;                                                 //  restart allowed                 v.13.04
   EFtonemap.threadfunc = Tmap_thread;
   if (! edit_setup(EFtonemap)) return;                                    //  setup: no preview, select area OK

/***
            _____________________________
           |                             |
           |                             |
           |    curve drawing area       |
           |                             |
           |_____________________________| 
            low       contrast       high

            Amplify ========[]=========

            Curve File: [ Open ] [ Save ]
***/

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFtonemap.zd = zd;

   zdialog_add_widget(zd,"frame","frame","dialog",0,"expand");
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labcL","hb1",ZTX("low"),"space=4");
   zdialog_add_widget(zd,"label","labcM","hb1",Bcontrast,"expand");
   zdialog_add_widget(zd,"label","labcH","hb1",ZTX("high"),"space=5");

   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labcon","hb2",ZTX("Amplify"),"space=5");
   zdialog_add_widget(zd,"hscale","amplify","hb2","0|1|0.005|0","expand");
   zdialog_add_widget(zd,"label","ampval","hb2",0,"space=5");

   zdialog_add_widget(zd,"hbox","hbcf","dialog",0,"space=8");
   zdialog_add_widget(zd,"label","labcf","hbcf",Bcurvefile,"space=5");
   zdialog_add_widget(zd,"button","loadcurve","hbcf",Bopen,"space=5");
   zdialog_add_widget(zd,"button","savecurve","hbcf",Bsave,"space=5");

   GtkWidget *frame = zdialog_widget(zd,"frame");                          //  set up curve edit
   spldat *sd = splcurve_init(frame,Tmap_curvedit);
   EFtonemap.curves = sd;

   sd->Nspc = 1;
   sd->vert[0] = 0;
   sd->nap[0] = 3;                                                         //  initial curve anchor points
   sd->apx[0][0] = 0.01;
   sd->apy[0][0] = 0.4;
   sd->apx[0][1] = 0.50;
   sd->apy[0][1] = 0.4;
   sd->apx[0][2] = 0.99;
   sd->apy[0][2] = 0.2;

   splcurve_generate(sd,0);                                                //  generate curve data

   Tmap_initz(zd);                                                         //  initialize                   v.13.04
   return;
}


//  initialization for new image file                                      //  v.13.04

void Tmap_initz(zdialog *zd)
{
   int         ii, cc, px, py;
   float       *pix1;
   int         jj, sum, limit, condist[100];
   
   if (Tmap_brmap1) {
      free(Tmap_brmap1);                                                   //  free memory
      for (int ii = 0; ii < 4; ii++)
      free(Tmap_brmap3[ii]);
   }

   cc = E1pxm->ww * E1pxm->hh * sizeof(float);                             //  allocate brightness map memory
   Tmap_brmap1 = (float *) malloc(cc);
   for (ii = 0; ii < 4; ii++)
      Tmap_brmap3[ii] = (float *) malloc(cc);

   for (py = 0; py < E1pxm->hh; py++)                                      //  map initial image brightness
   for (px = 0; px < E1pxm->ww; px++)
   {
      ii = py * E1pxm->ww + px;
      pix1 = PXMpix(E1pxm,px,py);
      Tmap_brmap1[ii] = pixbright(pix1);                                   //  pixel brightness 0-255.9
   }

   for (ii = 0; ii < 100; ii++) 
      condist[ii] = 0;

   for (py = 1; py < E1pxm->hh; py++)                                      //  map contrast distribution
   for (px = 1; px < E1pxm->ww; px++)
   {
      ii = py * E1pxm->ww + px;
      jj = 0.388 * fabsf(Tmap_brmap1[ii] - Tmap_brmap1[ii-1]);             //  horiz. contrast, ranged 0 - 99
      condist[jj]++;
      jj = 0.388 * fabsf(Tmap_brmap1[ii] - Tmap_brmap1[ii-E1pxm->ww]);     //  vertical
      condist[jj]++;
   }
   
   sum = 0;
   limit = 0.99 * 2 * (E1pxm->ww-1) * (E1pxm->hh-1);                       //  find 99th percentile contrast

   for (ii = 0; ii < 100; ii++) {
      sum += condist[ii];
      if (sum > limit) break;
   }
   
   Tmap_contrast99 = 255.0 * ii / 100.0;                                   //  0 to 255
   if (Tmap_contrast99 < 4) Tmap_contrast99 = 4;                           //  rescale low-contrast image

   zdialog_resize(zd,300,300);
   zdialog_run(zd,Tmap_dialog_event,"save");                               //  run dialog - parallel

   Tmap_amplify = 0;
   return;
}


//  dialog event and completion callback function

int Tmap_dialog_event(zdialog *zd, cchar *event)
{
   spldat   *sd = EFtonemap.curves;
   char     text[8];
   
   if (strEqu(event,"done")) zd->zstat = 1;                                //  apply and quit               v.13.12
   if (strEqu(event,"enter")) zd->zstat = 1;                               //  v.14.03

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done(0);
      else edit_cancel(0);
      free(Tmap_brmap1);                                                   //  free memory
      for (int ii = 0; ii < 4; ii++)
      free(Tmap_brmap3[ii]);
      Tmap_brmap1 = 0;                                                     //  v.13.04
      return 1;
   }

   if (strEqu(event,"blendwidth")) signal_thread();

   if (strEqu(event,"amplify")) {                                          //  slider value
      zdialog_fetch(zd,"amplify",Tmap_amplify);
      sprintf(text,"%.2f",Tmap_amplify);                                   //  numeric feedback
      zdialog_stuff(zd,"ampval",text);
      signal_thread();                                                     //  trigger update thread
   }

   if (strEqu(event,"loadcurve")) {                                        //  load saved curve 
      splcurve_load(sd);
      signal_thread();
      return 0;
   }

   if (strEqu(event,"savecurve")) {                                        //  save curve to file 
      splcurve_save(sd);
      return 0;
   }
   
   return 0;
}


//  this function is called when the curve is edited

void Tmap_curvedit(int)
{
   signal_thread();
   return;
}


//  thread function

void * Tmap_thread(void *)
{
   void * Tmap_wthread1(void *arg);
   void * Tmap_wthread2(void *arg);

   int      ii;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (ii = 0; ii < 4; ii++)                                           //  start working threads1 (must be 4)
         start_wthread(Tmap_wthread1,&Nval[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      for (ii = 0; ii < Nwt; ii++)                                         //  start working threads2
         start_wthread(Tmap_wthread2,&Nval[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      CEF->Fmods++; 
      CEF->Fsaved = 0;

      if (Tmap_amplify == 0)                                               //  v.13.04
         CEF->Fmods = 0;

      mwpaint2(); 
   }

   return 0;                                                               //  not executed, stop g++ warning
}


//  working threads

void * Tmap_wthread1(void *arg)
{
   int         ii, kk, bii, pii, dist = 0;
   int         px, px1, px2, pxinc;
   int         py, py1, py2, pyinc;
   float       b1, b3, xval, yval, grad;
   float       amplify, contrast99;
   spldat      *sd = EFtonemap.curves;

   contrast99 = Tmap_contrast99;                                           //  99th percentile contrast
   contrast99 = 1.0 / contrast99;                                          //  inverted

   amplify = pow(Tmap_amplify,0.4);                                        //  get amplification, 0 to 1    v.12.06

   bii = *((int *) arg);

   if (bii == 0) {                                                         //  direction SE
      px1 = 1; px2 = E3pxm->ww; pxinc = 1;
      py1 = 1; py2 = E3pxm->hh; pyinc = 1;
      pii = - 1 - E3pxm->ww;
   }
   
   else if (bii == 1) {                                                    //  direction SW
      px1 = E3pxm->ww-2; px2 = 0; pxinc = -1;
      py1 = 1; py2 = E3pxm->hh; pyinc = 1;
      pii = + 1 - E3pxm->ww;
   }
   
   else if (bii == 2) {                                                    //  direction NE
      px1 = 1; px2 = E3pxm->ww; pxinc = 1;
      py1 = E3pxm->hh-2; py2 = 0; pyinc = -1;
      pii = - 1 + E3pxm->ww;
   }
   
   else {   /* bii == 3 */                                                 //  direction NW
      px1 = E3pxm->ww-2; px2 = 0; pxinc = -1;
      py1 = E3pxm->hh-2; py2 = 0; pyinc = -1;
      pii = + 1 + E3pxm->ww;
   }

   for (ii = 0; ii < E3pxm->ww * E3pxm->hh; ii++)                          //  initial brightness map
      Tmap_brmap3[bii][ii] = Tmap_brmap1[ii];

   for (py = py1; py != py2; py += pyinc)                                  //  loop all image pixels
   for (px = px1; px != px2; px += pxinc)
   {
      ii = py * E3pxm->ww + px;

      if (sa_stat == 3) {                                                  //  select area active
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  outside pixel
      }

      b1 = Tmap_brmap1[ii];                                                //  this pixel brightness
      grad = b1 - Tmap_brmap1[ii+pii];                                     //   - prior pixel --> gradient

      xval = fabsf(grad) * contrast99;                                     //  gradient scaled 0 to 1+
      kk = 1000.0 * xval;
      if (kk > 999) kk = 999;
      yval = 1.0 + 5.0 * sd->yval[0][kk];

      grad = grad * yval;                                                  //  magnified gradient

      b3 = Tmap_brmap3[bii][ii+pii] + grad;                                //  pixel brightness = prior + gradient
      b3 = (1.0 - amplify) * b1 + amplify * b3;                            //  constrain: push b3 toward b1

      if (b3 > 255) b3 = 255;                                              //  constrain
      if (b3 < 1) b3 = 1;

      Tmap_brmap3[bii][ii] = b3;                                           //  new pixel brightness
   }
   
   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


void * Tmap_wthread2(void *arg)
{
   float       *pix1, *pix3;
   int         index, ii, px, py, dist = 0;
   float       b1, b3, bf, f1, f2;
   float       red1, green1, blue1, red3, green3, blue3;

   index = *((int *) arg);

   for (py = index; py < E3pxm->hh; py += Nwt)                             //  loop all image pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      ii = py * E3pxm->ww + px;

      if (sa_stat == 3) {                                                  //  select area active 
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel is outside area
      }

      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel

      b1 = Tmap_brmap1[ii];                                                //  initial pixel brightness
      b3 = Tmap_brmap3[0][ii] + Tmap_brmap3[1][ii]                         //  new brightness = average of four
         + Tmap_brmap3[2][ii] + Tmap_brmap3[3][ii];                        //    calculated brightness surfaces

      bf = 0.25 * b3 / (b1 + 1);                                           //  brightness ratio
      
      red1 = pix1[0];                                                      //  input RGB
      green1 = pix1[1];
      blue1 = pix1[2];
      
      red3 = bf * red1;                                                    //  output RGB
      if (red3 > 255.9) red3 = 255.9;
      green3 = bf * green1;
      if (green3 > 255.9) green3 = 255.9;
      blue3 = bf * blue1;
      if (blue3 > 255.9) blue3 = 255.9;

      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         red3 = f1 * red3 + f2 * red1;
         green3 = f1 * green3 + f2 * green1;
         blue3 = f1 * blue3 + f2 * blue1;
      }
      
      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  adjust brightness distribution by flattening and/or expanding range

namespace brightdist_names
{
   float    flatten, deband;                                               //  flatten/deband values 0 - 100%
   float    darken, brighten;                                              //  darken/brighten values, 0 - 100%
   float    flatten_dist[1000];                                            //  calculated flattened distribution

   int    dialog_event(zdialog* zd, cchar *event);
   void * thread(void *);
   void   flatten_func();
   void * flatten_wthread(void *);
   void   expand_func();

   editfunc    EFbrightdist;
}


//  menu function

void m_brightdist(GtkWidget *, cchar *)                                    //  combine flatten and expand      v.13.10
{
   using namespace brightdist_names;

   cchar  *title = ZTX("Adjust Brightness Distribution");
   F1_help_topic = "brightness_distribution";

   EFbrightdist.menufunc = m_brightdist;
   EFbrightdist.funcname = "brightdist";
   EFbrightdist.FprevReq = 1;                                              //  preview
   EFbrightdist.Farea = 2;                                                 //  select area usable
   EFbrightdist.Frestart = 1;                                              //  restart allowed
   EFbrightdist.threadfunc = thread;
   if (! edit_setup(EFbrightdist)) return;                                 //  setup edit

/***
          ______________________________________
         |    Adjust Brightness Distribution    |
         |                                      |
         | Flatten  =========[]============ NN  |
         |  Deband  ============[]========= NN  |
         |                                      |
         |  Darken  ===[]================== NN  |
         | Brighten ==[]=================== NN  |
         |                                      |
         |                     [Done] [Cancel]  |
         |______________________________________|

***/

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFbrightdist.zd = zd;
   
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb11","hb1",0,"homog|space=3");
   zdialog_add_widget(zd,"vbox","vb12","hb1",0,"homog|expand");
   zdialog_add_widget(zd,"vbox","vb13","hb1",0,"homog|space=3");
   zdialog_add_widget(zd,"vbox","vb21","hb2",0,"homog|space=3");
   zdialog_add_widget(zd,"vbox","vb22","hb2",0,"homog|expand");
   zdialog_add_widget(zd,"vbox","vb23","hb2",0,"homog|space=3");

   zdialog_add_widget(zd,"label","labflatten","vb11",ZTX("Flatten"));
   zdialog_add_widget(zd,"label","labdeband","vb11",ZTX("Deband"));
   zdialog_add_widget(zd,"label","labdarken","vb21",ZTX("Darken"));
   zdialog_add_widget(zd,"label","labbrighten","vb21",ZTX("Brighten"));

   zdialog_add_widget(zd,"hscale","flatten","vb12","0|100|1|0","expand");
   zdialog_add_widget(zd,"hscale","deband","vb12","0|100|1|0","expand");
   zdialog_add_widget(zd,"hscale","darken","vb22","0|100|0.2|0","expand");
   zdialog_add_widget(zd,"hscale","brighten","vb22","0|100|0.2|0","expand");

   zdialog_add_widget(zd,"label","flattenval","vb13","0");
   zdialog_add_widget(zd,"label","debandval","vb13","0");
   zdialog_add_widget(zd,"label","darkenval","vb23","0");
   zdialog_add_widget(zd,"label","brightenval","vb23","0");

   zdialog_resize(zd,300,0);
   zdialog_run(zd,dialog_event,"save");                                    //  run dialog - parallel

   flatten = deband = 0;                                                   //  initz. parameters
   darken = brighten = 0;

   m_histogram(0,0);                                                       //  popup brightness histogram
   return;
}


//  dialog event and completion function

int brightdist_names::dialog_event(zdialog *zd, cchar *event)
{
   using namespace brightdist_names;

   char     text[8];

   if (strEqu(event,"done")) zd->zstat = 1;                                //  apply and quit               v.13.12
   if (strEqu(event,"enter")) zd->zstat = 1;                               //  v.14.03

   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done(0);                                    //  done
      else edit_cancel(0);                                                 //  cancel or destroy
      m_histogram(0,"kill");
      return 1;
   }

   if (strEqu(event,"blendwidth"))                                         //  select area blendwidth change
      signal_thread();

   if (strEqu(event,"flatten")) {
      zdialog_fetch(zd,"flatten",flatten);                                 //  get slider values
      sprintf(text,"%.1f",flatten);                                        //    and update numeric feedback 
      zdialog_stuff(zd,"flattenval",text);
      signal_thread();
   }
   
   if (strEqu(event,"deband")) {
      zdialog_fetch(zd,"deband",deband);
      sprintf(text,"%.1f",deband);
      zdialog_stuff(zd,"debandval",text);
      signal_thread();
   }

   if (strEqu(event,"darken")) {
      zdialog_fetch(zd,"darken",darken);
      sprintf(text,"%.1f",darken);
      zdialog_stuff(zd,"darkenval",text);
      signal_thread();
   }

   if (strEqu(event,"brighten")) {
      zdialog_fetch(zd,"brighten",brighten);
      sprintf(text,"%.1f",brighten);
      zdialog_stuff(zd,"brightenval",text);
      signal_thread();
   }

   wait_thread_idle();
   mwpaintnow();
   zmainloop();
   return 1;
}


//  brightness distribution thread function

void * brightdist_names::thread(void *)
{
   using namespace brightdist_names;

   int         px, py, ii, npix, ww, hh;
   float       bright1;
   float       *pix1;

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      ww = E1pxm->ww;                                                      //  bugfix  v.13.11.1
      hh = E1pxm->hh;
      
      for (ii = 0; ii < 1000; ii++)                                        //  clear brightness distribution data
         flatten_dist[ii] = 0;

      if (sa_stat == 3)                                                    //  process selected area
      {
         for (ii = npix = 0; ii < ww * hh; ii++)
         {
            if (! sa_pixmap[ii]) continue;
            py = ii / ww;
            px = ii - py * ww;
            pix1 = PXMpix(E1pxm,px,py);
            bright1 = 1000.0 / 256.0 * pixbright(pix1);
            flatten_dist[int(bright1)]++;
            npix++;
         }
         
         for (ii = 1; ii < 1000; ii++)                                     //  cumulative brightness distribution
            flatten_dist[ii] += flatten_dist[ii-1];                        //   0 ... npix

         for (ii = 0; ii < 1000; ii++)
            flatten_dist[ii] = flatten_dist[ii]                            //  multiplier per brightness level
                             / npix * 1000.0 / (ii + 1);
      }

      else                                                                 //  process whole image
      {
         for (py = 0; py < hh; py++)                                       //  compute brightness distribution
         for (px = 0; px < ww; px++)
         {
            pix1 = PXMpix(E1pxm,px,py);
            bright1 = 1000.0 / 256.0 * pixbright(pix1);
            flatten_dist[int(bright1)]++;
         }
         
         for (ii = 1; ii < 1000; ii++)                                     //  cumulative brightness distribution
            flatten_dist[ii] += flatten_dist[ii-1];                        //   0 ... (ww * hh)

         for (ii = 0; ii < 1000; ii++)
            flatten_dist[ii] = flatten_dist[ii]                            //  multiplier per brightness level 
                             / (ww * hh) * 1000.0 / (ii + 1);
      }
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker thread per processor core
         start_wthread(flatten_wthread,&Nval[ii]);
      wait_wthreads();                                                     //  wait for completion

      if (! flatten && ! darken && ! brighten)                             //  no modifications
         CEF->Fmods = 0;
      else {
         CEF->Fmods++;                                                     //  image modified, not saved
         CEF->Fsaved = 0;
      }
   }

   return 0;                                                               //  not executed, stop g++ warning
}


//  worker thread for each CPU processor core

void * brightdist_names::flatten_wthread(void *arg)
{
   using namespace brightdist_names;

   int         index = *((int *) (arg));
   int         px, py, ii, dist = 0;
   float       *pix1, *pix3;
   float       fold, fnew, dold, dnew, cmax;
   float       red1, green1, blue1, red3, green3, blue3;
   float       bright1, bright2, deband2;
   float       dark, bright, b1, b3, bf, f1, f2;
   
   //  flatten brightness distribution

   for (py = index; py < E1pxm->hh; py += Nwt)                             //  flatten brightness distribution
   for (px = 0; px < E1pxm->ww; px++)
   {
      if (sa_stat == 3) {                                                  //  select area active
         ii = py * E1pxm->ww + px;   
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  outside pixel
      }

      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel
         
      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];

      bright1 = 0.25 * red1 + 0.65 * green1 + 0.10 * blue1;                //  input pixel brightness 0-255.9
      bright1 *= 1000.0 / 256.0;                                           //  0-999
      bright2 = flatten_dist[int(bright1)];                                //  output adjustment factor

      red3 = bright2 * red1;                                               //  fully flattened brightness
      green3 = bright2 * green1;
      blue3 = bright2 * blue1;

      deband2 = (1000.0 - bright1) / 1000.0;                               //  bright1 = 0 to 1000  >>  deband2 = 1 to 0
      deband2 += (1.0 - deband2) * (1.0 - 0.01 * deband);                  //  deband2 = 1 to 1  >>  1 to 0

      fnew = 0.01 * flatten;                                               //  how much to flatten, 0 to 1
      fnew = deband2 * fnew;                                               //  attenuate flatten of brighter pixels
      fold = 1.0 - fnew;                                                   //  how much to retain, 1 to 0

      red3 = fnew * red3 + fold * red1;                                    //  blend new and old brightness
      green3 = fnew * green3 + fold * green1;
      blue3 = fnew * blue3 + fold * blue1;

      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         dnew = 1.0 * dist / sa_blend;                                     //    blend changes over sa_blend
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }

      cmax = red3;                                                         //  stop overflow, keep color balance
      if (green3 > cmax) cmax = green3;
      if (blue3 > cmax) cmax = blue3;
      if (cmax > 255.9) {
         cmax = 255.9 / cmax;
         red3 = red3 * cmax;
         green3 = green3 * cmax;
         blue3 = blue3 * cmax;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }
   
   //  expand brightness distribution

   dark = 0.01 * darken * 255.9;                                           //  clipping brightness levels
   bright = (1.0 - 0.01 * brighten) * 255.9;                               //  dark = 0-255.9  bright = 255.9-0
   
   if (dark >= bright) {                                                   //  inputs mean 100% clipping
      dark = bright - 0.01;                                                //    retain 1-2% of range          v.13.10
      bright += 0.01;
      if (dark < 0) dark = 0;
      if (bright > 255.9) bright = 255.9;
   }

   for (py = index; py < E1pxm->hh; py += Nwt)                             //  expand brightness distribution
   for (px = 0; px < E1pxm->ww; px++)
   {
      ii = py * E3pxm->ww + px;

      if (sa_stat == 3) {                                                  //  select area active
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel is outside area
      }

      pix1 = pix3 = PXMpix(E3pxm,px,py);                                   //  input and output pixel          v.13.10

      b1 = pixbright(pix1);                                                //  0-255.9

      if (b1 < dark)                                                       //  clip dark pixels
         b3 = 0;
      else if (b1 > bright)                                                //  clip bright pixels
         b3 = bright;
      else b3 = b1;
      
      if (b3 > dark) b3 = b3 - dark;                                       //  expand the rest
      b3 = b3 * (255.9 / (bright - dark));

      bf = b3 / (b1 + 0.1);                                                //  brightness ratio
      
      red1 = pix1[0];                                                      //  input RGB 
      green1 = pix1[1];
      blue1 = pix1[2];
      
      red3 = bf * red1;                                                    //  output RGB
      green3 = bf * green1;
      blue3 = bf * blue1;
      
      if (red3 > 255.9) red3 = 255.9;
      if (green3 > 255.9) green3 = 255.9;
      if (blue3 > 255.9) blue3 = 255.9;
      
      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         red3 = f1 * red3 + f2 * red1;
         green3 = f1 * green3 + f2 * green1;
         blue3 = f1 * blue3 + f2 * blue1;
      }
      
      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  ramp brightness across image, vertical and horizontal gradients

editfunc    EFbrramp;
int         brramp_spc;                                                    //  current spline curves (2 at a time)

void m_brightramp(GtkWidget *, cchar *)                                    //  upgrade for 4 colors
{
   int    brramp_dialog_event(zdialog* zd, cchar *event);
   void   brramp_curvedit(int spc);
   void * brramp_thread(void *);
   void   brramp_init();

   F1_help_topic = "brightness_ramp";

   EFbrramp.menufunc = m_brightramp;
   EFbrramp.funcname = "bright-ramp";
   EFbrramp.FprevReq = 1;                                                  //  use preview
   EFbrramp.Farea = 2;                                                     //  select area usable
   EFbrramp.threadfunc = brramp_thread;
   if (! edit_setup(EFbrramp)) return;                                     //  setup edit
   
/***
          _________________________________________
         |                                         |
         |                                         |
         |           curve edit area               |
         |                                         |
         |_________________________________________|

            (o) all  (o) red  (o) green  (o) blue                          //  colors: all, red, green, blue
            Curve File: [Open]  [Save]                                     //  each has vert. and horiz. curve
                                   [Done]  [Cancel]    

***/

   zdialog *zd = zdialog_new(ZTX("Ramp brightness across image"),mWin,Bdone,Bcancel,null);
   EFbrramp.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5|expand");
   zdialog_add_widget(zd,"vbox","vb1","hb1");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"expand");
   zdialog_add_widget(zd,"label","lmax","vb1","+","space=3");
   zdialog_add_widget(zd,"label","lspace","vb1",0,"expand");
   zdialog_add_widget(zd,"label","lmin","vb1","‒","space=3");
   zdialog_add_widget(zd,"label","lspace","vb1");
   zdialog_add_widget(zd,"frame","frame","vb2",0,"expand");

   zdialog_add_widget(zd,"hbox","hb2","vb2");
   zdialog_add_widget(zd,"label","lmin","hb2","‒","space=3");
   zdialog_add_widget(zd,"label","lspace","hb2",0,"expand");
   zdialog_add_widget(zd,"label","lmax","hb2","+","space=3");
   zdialog_add_widget(zd,"hbox","hbrgb","dialog",0,"space=3");
   zdialog_add_widget(zd,"radio","all","hbrgb",Ball,"space=5");
   zdialog_add_widget(zd,"radio","red","hbrgb",Bred,"space=5");
   zdialog_add_widget(zd,"radio","green","hbrgb",Bgreen,"space=5");
   zdialog_add_widget(zd,"radio","blue","hbrgb",Bblue,"space=5");
   zdialog_add_widget(zd,"hbox","hbcf","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcf","hbcf",Bcurvefile,"space=8");
   zdialog_add_widget(zd,"button","loadcurves","hbcf",Bopen,"space=5");
   zdialog_add_widget(zd,"button","savecurves","hbcf",Bsave,"space=5");

   GtkWidget *frame = zdialog_widget(zd,"frame");
   spldat *sd = splcurve_init(frame,brramp_curvedit);
   EFbrramp.curves = sd;
   
   brramp_init();                                                          //  set up initial flat curves

   sd->Nspc = 2;                                                           //  only curves 0-1 are shown
   brramp_spc = 2;                                                         //  current curves: all (2-3)
   
   zdialog_stuff(zd,"all",1);                                              //  stuff default selection, all

   zdialog_resize(zd,260,350);
   zdialog_run(zd,brramp_dialog_event,"save");                             //  run dialog - parallel
   return;
}


//  initialize all curves to flat

void brramp_init()
{
   spldat *sd = EFbrramp.curves;

   for (int ii = 0; ii < 10; ii++)                                         //  loop curves: current (0-1), all (2-3),
   {                                                                       //    red (4-5), green (6-7), blue (8-9)
      sd->nap[ii] = 2;                                                     //  horizontal curve ii
      sd->vert[ii] = 0;
      sd->apx[ii][0] = 0.01;
      sd->apx[ii][1] = 0.99;
      sd->apy[ii][0] = sd->apy[ii][1] = 0.5;
      splcurve_generate(sd,ii);

      ii++;
      sd->nap[ii] = 2;                                                     //  vertical curve ii+1
      sd->vert[ii] = 1;
      sd->apx[ii][0] = 0.01;
      sd->apx[ii][1] = 0.99;
      sd->apy[ii][0] = sd->apy[ii][1] = 0.5;
      splcurve_generate(sd,ii);
   }
   
   return;
}


//  dialog event and completion callback function

int brramp_dialog_event(zdialog *zd, cchar *event)
{
   spldat      *sd = EFbrramp.curves;
   int         ii, jj, nn, Fupdate = 0;
   spldat      sdtemp;

   if (strEqu(event,"done")) zd->zstat = 1;                                //  apply and quit               v.13.12
   if (strEqu(event,"enter")) zd->zstat = 1;                               //  v.14.03

   if (zd->zstat) {                                                        //  dialog complete
      if (zd->zstat == 1) edit_done(0);
      else edit_cancel(0);
      return 1;
   }

   if (strEqu(event,"blendwidth")) signal_thread();

   if (strstr("all red green blue",event)) {                               //  new choice of curve
      zdialog_fetch(zd,event,ii);
      if (! ii) return 0;                                                  //  ignore button OFF events

      ii = strcmpv(event,"all","red","green","blue",null);
      brramp_spc = ii = 2 * ii;                                            //  2, 4, 6, 8

      sd->nap[0] = sd->nap[ii];                                            //  copy active curves ii/ii+1 to curves 0/1
      sd->nap[1] = sd->nap[ii+1];
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }
      for (jj = 0; jj < sd->nap[1]; jj++) {
         sd->apx[1][jj] = sd->apx[ii+1][jj];
         sd->apy[1][jj] = sd->apy[ii+1][jj];
      }

      Fupdate++;
   }

   if (strEqu(event,"loadcurves"))                                         //  load 8 saved curves from file
   {
      sdtemp.Nspc = 8;
      sdtemp.drawarea = 0;
      nn = splcurve_load(&sdtemp);
      if (nn != 1) return 0;

      for (ii = 0; ii < 8; ii++) {                                         //  load curves 0-7 to curves 2-9
         sd->vert[ii+2] = sdtemp.vert[ii];
         sd->nap[ii+2] = sdtemp.nap[ii];
         for (jj = 0; jj < sdtemp.nap[ii]; jj++) {
            sd->apx[ii+2][jj] = sdtemp.apx[ii][jj];
            sd->apy[ii+2][jj] = sdtemp.apy[ii][jj];
         }
         splcurve_generate(sd,ii+2);
      }

      ii = brramp_spc;                                                     //  current active curves: 2, 4, 6, 8

      sd->nap[0] = sd->nap[ii];                                            //  copy 2 active curves to curves 0-1
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }

      sd->nap[1] = sd->nap[ii+1];
      for (jj = 0; jj < sd->nap[1]; jj++) {
         sd->apx[1][jj] = sd->apx[ii+1][jj];
         sd->apy[1][jj] = sd->apy[ii+1][jj];
      }

      Fupdate++;
   }

   if (strEqu(event,"savecurves"))                                         //  save 8 curves to file
   {
      sdtemp.Nspc = 8;
      for (ii = 0; ii < 8; ii++) {
         sdtemp.vert[ii] = sd->vert[ii+2];
         sdtemp.nap[ii] = sd->nap[ii+2];
         for (jj = 0; jj < sd->nap[ii+2]; jj++) {
            sdtemp.apx[ii][jj] = sd->apx[ii+2][jj];
            sdtemp.apy[ii][jj] = sd->apy[ii+2][jj];
         }
      }

      splcurve_save(&sdtemp);
   }

   if (Fupdate)                                                            //  curve has changed
   {
      splcurve_generate(sd,0);                                             //  regenerate curves 0-1
      splcurve_generate(sd,1);
      
      ii = brramp_spc;                                                     //  active curves

      sd->nap[ii] = sd->nap[0];                                            //  copy curves 0-1 to active curves
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[ii][jj] = sd->apx[0][jj];
         sd->apy[ii][jj] = sd->apy[0][jj];
      }

      sd->nap[ii+1] = sd->nap[1];
      for (jj = 0; jj < sd->nap[1]; jj++) {
         sd->apx[ii+1][jj] = sd->apx[1][jj];
         sd->apy[ii+1][jj] = sd->apy[1][jj];
      }

      for (jj = 0; jj < 1000; jj++) 
      {
         sd->yval[ii][jj] = sd->yval[0][jj];
         sd->yval[ii+1][jj] = sd->yval[1][jj];
      }

      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve

      signal_thread();                                                     //  trigger image update
   }

   return 1;
}


//  this function is called when a curve is edited

void brramp_curvedit(int spc)
{
   int      ii = brramp_spc, jj;
   spldat   *sd = EFbrramp.curves;                                         //  active curves: 2, 4, 6, 8 (and +1)

   sd->nap[ii] = sd->nap[0];                                               //  copy curves 0-1 to active curves
   for (jj = 0; jj < sd->nap[0]; jj++) {
      sd->apx[ii][jj] = sd->apx[0][jj];
      sd->apy[ii][jj] = sd->apy[0][jj];
   }

   sd->nap[ii+1] = sd->nap[1];
   for (jj = 0; jj < sd->nap[1]; jj++) {
      sd->apx[ii+1][jj] = sd->apx[1][jj];
      sd->apy[ii+1][jj] = sd->apy[1][jj];
   }

   for (jj = 0; jj < 1000; jj++) {
      sd->yval[ii][jj] = sd->yval[0][jj];
      sd->yval[ii+1][jj] = sd->yval[1][jj];
   }

   if (brramp_spc == 2) {                                                  //  "all" curve was edited
      for (ii = 4; ii < 10; ii += 2) {                                     //  fix R/G/B curves to match
         sd->nap[ii] = sd->nap[2];
         for (jj = 0; jj < sd->nap[2]; jj++) {
            sd->apx[ii][jj] = sd->apx[2][jj];
            sd->apy[ii][jj] = sd->apy[2][jj];
         }
         for (jj = 0; jj < 1000; jj++)
            sd->yval[ii][jj] = sd->yval[2][jj];
      }
      for (ii = 5; ii < 10; ii += 2) {
         sd->nap[ii] = sd->nap[3];
         for (jj = 0; jj < sd->nap[3]; jj++) {
            sd->apx[ii][jj] = sd->apx[3][jj];
            sd->apy[ii][jj] = sd->apy[3][jj];
         }
         for (jj = 0; jj < 1000; jj++)
            sd->yval[ii][jj] = sd->yval[3][jj];
      }
   }

   signal_thread();
   return;
}


//  brramp thread function

void * brramp_thread(void *arg)
{
   void * brramp_wthread(void *);
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(brramp_wthread,&Nval[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      mwpaint2();                                                          //  update window
      CEF->Fmods++;                                                        //  image3 modified
      CEF->Fsaved = 0;
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * brramp_wthread(void *arg)                                           //  worker thread function
{
   int         index = *((int *) arg);
   int         ii, jj, kk, dist = 0, px3, py3;
   float       *pix1, *pix3;
   float       red1, green1, blue1, maxrgb;
   float       red3, green3, blue3;
   float       dispx, dispy;
   float       hramp[3], vramp[3], tramp[3];
   float       spanw, spanh, dold, dnew;
   spldat      *sd = EFbrramp.curves;

   if (sa_stat == 3) {                                                     //  if select area active, ramp
      spanw = sa_maxx - sa_minx;                                           //    brightness over enclosing rectangle
      spanh = sa_maxy - sa_miny;
   }
   else {
      spanw = E3pxm->ww;                                                   //  else over entire image
      spanh = E3pxm->hh;
   }   

   for (py3 = index; py3 < E3pxm->hh; py3 += Nwt)                          //  loop output pixels
   for (px3 = 0; px3 < E3pxm->ww; px3++)
   {
      if (sa_stat == 3) {                                                  //  select area active 
         ii = py3 * E3pxm->ww + px3;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
         dispx = (px3 - sa_minx) / spanw;
         dispy = (py3 - sa_miny) / spanh;
      }
      else {
         dispx = px3 / spanw;                                              //  left > right = 0 to 1
         dispy = py3 / spanh;                                              //  top > bottom = 0 to 1
      }
      
      kk = 1000 * dispx;                                                   //  conv. dispx 0-1 to 0-1000
      if (kk > 999) kk = 999;

      for (ii = 0, jj = 4; ii < 3; ii++, jj+=2)                            //  horz. curves for dispx and R/G/B
         hramp[ii] = sd->yval[jj][kk] - 0.5;                               //  scale to -0.5 to +0.5
      
      kk = 1000 * dispy;                                                   //  conv. dispy
      if (kk > 999) kk = 999;

      for (ii = 0, jj = 5; ii < 3; ii++, jj+=2)                            //  vert. curves
         vramp[ii] = sd->yval[jj][kk] - 0.5;

      for (ii = 0; ii < 3; ii++)                                           //  total R/G/B change
         tramp[ii] = 1.0 + hramp[ii] + vramp[ii];                          //  scale to 0.0 to 2.0

      pix1 = PXMpix(E1pxm,px3,py3);                                        //  input pixel
      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];
      
      red3 = red1 * tramp[0];                                              //  change R/G/B
      green3 = green1 * tramp[1];
      blue3 = blue1 * tramp[2];

      maxrgb = red3;
      if (green3 > maxrgb) maxrgb = green3;
      if (blue3 > maxrgb) maxrgb = blue3;
      
      if (maxrgb > 255.9) {                                                //  stop overflow
         red3 = red3 * 255.9 / maxrgb;
         green3 = green3 * 255.9 / maxrgb;
         blue3 = blue3 * 255.9 / maxrgb;
      }

      if (sa_stat == 3 && dist < sa_blend) {                               //  blend changes over blendwidth
         dnew = 1.0 * dist / sa_blend;
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }

      pix3 = PXMpix(E3pxm,px3,py3);                                        //  output pixel
      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  Select area and edit in parallel
//  Current edit function is applied to areas painted with the mouse.
//  Mouse can be weak or strong, and edits are applied incrementally.
//  method: 
//    entire image is a select area with all pixel edge distance = 0 (outside area)
//    blendwidth = 10000 (infinite)
//    pixels painted with mouse have increasing edge distance to amplify edits

int   paint_edits_radius;
int   paint_edits_cpower;
int   paint_edits_epower;

void m_paint_edits(GtkWidget *, cchar *)                                   //  menu function
{
   int   paint_edits_dialog_event(zdialog *, cchar *event);                //  dialog event function
   void  paint_edits_mousefunc();                                          //  mouse function

   cchar    *title = ZTX("Paint Edits");
   cchar    *helptext1 = ZTX("Press F1 for help");
   cchar    *helptext2 = ZTX("Edit function must be active");
   int      yn, cc;

   F1_help_topic = "paint_edits";

   if (! curr_file) return;                                                //  no image
   
   if (sa_stat) {
      yn = zmessageYN(mWin,ZTX("Select area cannot be kept.\n"             //  v.13.04
                               "Continue?"));
      if (! yn) return;
      sa_unselect();                                                       //  unselect area
      zdialog_free(zdsela);
   }

   if (CEF && CEF->Fpreview) edit_fullsize();                              //  use full-size image
   
/***
    ____________________________________________
   |           Press F1 for help                |
   |       Edit Function must be active         |
   | mouse radius [___|v]                       |
   | power:  center [___|v]  edge [___|v]       |
   | [reset area]                               |
   |                                  [done]    |
   |____________________________________________|
   
***/

   zdsela = zdialog_new(title,mWin,Bdone,null);
   zdialog_add_widget(zdsela,"label","labhelp1","dialog",helptext1,"space=5");
   zdialog_add_widget(zdsela,"label","labhelp2","dialog",helptext2);
   zdialog_add_widget(zdsela,"label","labspace","dialog");
   zdialog_add_widget(zdsela,"hbox","hbr","dialog",0,"space=3");
   zdialog_add_widget(zdsela,"label","labr","hbr",ZTX("mouse radius"),"space=5");
   zdialog_add_widget(zdsela,"spin","radius","hbr","2|500|1|50");
   zdialog_add_widget(zdsela,"hbox","hbt","dialog",0,"space=3");
   zdialog_add_widget(zdsela,"label","labtc","hbt",ZTX("power:  center"),"space=5");
   zdialog_add_widget(zdsela,"spin","center","hbt","0|100|1|50");
   zdialog_add_widget(zdsela,"label","labte","hbt",Bedge,"space=5");
   zdialog_add_widget(zdsela,"spin","edge","hbt","0|100|1|0");
   zdialog_add_widget(zdsela,"hbox","hbra","dialog",0,"space=5");
   zdialog_add_widget(zdsela,"button","reset","hbra",ZTX("reset area"),"space=5");
   
   paint_edits_radius = 50;
   paint_edits_cpower = 50;
   paint_edits_epower = 0;

   cc = Fpxb->ww * Fpxb->hh * sizeof(uint16);                              //  allocate sa_pixmap[] for new area
   sa_pixmap = (uint16 *) malloc(cc);
   memset(sa_pixmap,0,cc);                                                 //  edge distance = 0 for all pixels

   sa_minx = 0;                                                            //  enclosing rectangle
   sa_maxx = Fpxb->ww;
   sa_miny = 0;
   sa_maxy = Fpxb->hh;

   sa_Npixel = Fpxb->ww * Fpxb->hh;
   sa_stat = 3;                                                            //  area status = complete
   sa_mode = mode_image;                                                   //  area mode = whole image
   sa_calced = 1;                                                          //  edge calculation complete
   sa_blend = 10000;                                                       //  "blend width"
   sa_fww = Fpxb->ww;                                                      //  valid image dimensions
   sa_fhh = Fpxb->hh;
   areanumber++;                                                           //  next sequential number

   zdialog_run(zdsela,paint_edits_dialog_event,"save");                    //  run dialog - parallel
   return;
}


//  Adjust whole image area to increase edit power for pixels within the mouse radius
//  sa_pixmap[*]  = 0 = never touched by mouse
//                = 1 = minimum edit power (barely painted)
//                = sa_blend = maximum edit power (edit fully applied)

int paint_edits_dialog_event(zdialog *zd, cchar *event)
{
   void  paint_edits_mousefunc();                                          //  mouse function
   int      cc;
   
   if (zd->zstat)                                                          //  done or cancel
   {
      freeMouse();                                                         //  disconnect mouse function
      zdialog_free(zdsela);                                                //  kill dialog
      sa_unselect();                                                       //  unselect area
      return 0;
   }

   if (sa_stat != 3) return 1;                                             //  area gone
   if (! sa_validate()) return 1;                                          //  area invalid for curr. image file  v.12.12

   if (strEqu(event,"focus")) {                                            //  toggle mouse capture
      if (CEF) takeMouse(paint_edits_mousefunc,0);
      else freeMouse();                                                    //  disconnect mouse
   }

   if (strEqu(event,"radius"))
      zdialog_fetch(zd,"radius",paint_edits_radius);                       //  set mouse radius

   if (strEqu(event,"center"))
      zdialog_fetch(zd,"center",paint_edits_cpower);                       //  set mouse center power

   if (strEqu(event,"edge"))
      zdialog_fetch(zd,"edge",paint_edits_epower);                         //  set mouse edge power
   
   if (strEqu(event,"reset")) {
      sa_unselect();                                                       //  unselect current area if any
      cc = Fpxb->ww * Fpxb->hh * sizeof(uint16);                           //  allocate sa_pixmap[] for new area
      sa_pixmap = (uint16 *) malloc(cc);
      memset(sa_pixmap,0,cc);                                              //  edge distance = 0 for all pixels

      sa_minx = 0;                                                         //  enclosing rectangle
      sa_maxx = Fpxb->ww;
      sa_miny = 0;
      sa_maxy = Fpxb->hh;

      sa_Npixel = Fpxb->ww * Fpxb->hh;
      sa_stat = 3;                                                         //  area status = complete
      sa_mode = mode_image;                                                //  area mode = whole image
      sa_calced = 1;                                                       //  edge calculation complete
      sa_blend = 10000;                                                    //  "blend width"
      sa_fww = Fpxb->ww;                                                   //  valid image dimensions
      sa_fhh = Fpxb->hh;
      areanumber++;                                                        //  next sequential number
   }
   
   return 1;
}


//  mouse function - adjust edit strength for areas within mouse radius
//  "edge distance" is increased for more strength, decreased for less

void paint_edits_mousefunc()
{
   int      ii, px, py, rx, ry, mrect;
   int      radius, radius2, cpower, epower;
   float    rad, rad2, power;

   if (! CEF) return;                                                      //  no active edit
   if (sa_stat != 3) return;                                               //  area gone?

   radius = paint_edits_radius;                                            //  pixel selection radius
   radius2 = radius * radius;
   cpower = paint_edits_cpower;
   epower = paint_edits_epower;

   mrect = 2 * radius;
   draw_mousearc(Mxposn,Myposn,mrect,mrect,0);                             //  show mouse selection circle

   if (LMclick || RMclick)                                                 //  mouse click, process normally
      return;

   if (Mbutton != 1 && Mbutton != 3)                                       //  button released
      return;
   
   Mxdrag = Mydrag = 0;                                                    //  neutralize drag                 v.12.11

   for (rx = -radius; rx <= radius; rx++)                                  //  loop every pixel in radius
   for (ry = -radius; ry <= radius; ry++)
   {
      rad2 = rx * rx + ry * ry;
      if (rad2 > radius2) continue;                                        //  outside radius
      px = Mxposn + rx;
      py = Myposn + ry;
      if (px < 0 || px > E1pxm->ww-1) continue;                            //  off the image edge
      if (py < 0 || py > E1pxm->hh-1) continue;

      ii = E1pxm->ww * py + px;
      rad = sqrt(rad2);
      power = cpower + rad / radius * (epower - cpower);                   //  power at pixel radius
      
      if (Mbutton == 1)                                                    //  left mouse button
      {                                                                    //  increase edit power
         sa_pixmap[ii] += power;
         if (sa_pixmap[ii] > sa_blend) sa_pixmap[ii] = sa_blend;
      }

      if (Mbutton == 3)                                                    //  right mouse button
      {                                                                    //  weaken edit power
         if (sa_pixmap[ii] <= power) sa_pixmap[ii] = 0;
         else sa_pixmap[ii] -= power;
      }
   }

   sa_minx = Mxposn - radius;                                              //  set temp. smaller area around mouse
   if (sa_minx < 0) sa_minx = 0;
   sa_maxx = Mxposn + radius;
   if (sa_maxx > E1pxm->ww) sa_maxx = E1pxm->ww;
   sa_miny = Myposn - radius;
   if (sa_miny < 0) sa_miny = 0;
   sa_maxy = Myposn + radius;
   if (sa_maxy > E1pxm->hh) sa_maxy = E1pxm->hh;
   sa_Npixel = (sa_maxx - sa_minx) * (sa_maxy - sa_miny);

   zdialog_send_event(CEF->zd,"blendwidth");                               //  notify edit dialog

   draw_mousearc(Mxposn,Myposn,mrect,mrect,0);                             //  show mouse selection circle

   sa_minx = 0;                                                            //  restore whole image area 
   sa_maxx = E1pxm->ww;
   sa_miny = 0;
   sa_maxy = E1pxm->hh;
   sa_Npixel = E1pxm->ww * E1pxm->hh;
   
   return;
}



